# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2019 Kostantinos Papadamou, Cyprus University of Technology
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# If you make use of this code in your work, please cite the following paper:
# @inproceedings{zannettou2018good,
#   title={The good, the bad and the bait: Detecting and characterizing clickbait on YouTube},
#   author={Zannettou, Savvas and Chatzis, Sotirios and Papadamou, Kostantinos and Sirivianos, Michael},
#   booktitle={2018 IEEE Security and Privacy Workshops (SPW)},
#   pages={63--69},
#   year={2018},
#   organization={IEEE}
# }

# Import Packages
from keras import backend as K
import tensorflow as tf

import numpy
import pickle
from tqdm import tqdm

from keras.layers import Input, Dense, Lambda, Dropout, Concatenate, Convolution2D, Reshape, MaxPooling2D
from keras.models import Model
from keras.optimizers import RMSprop
from keras import objectives

from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score

from sklearn.utils import shuffle
import data_preparation


"""
Initialize Parameters
"""
num_examples = 200000

"""
[DATA PREPARATION] Get all required data for Training and Testing the Model (Preload them to memory for faster training)
"""
# Get Dataset Video IDs
dataset_video_ids = pickle.load(open("all_video_ids_200k.p", "rb"))
dataset_video_ids = dataset_video_ids[:num_examples]

# Get VIDEO TITLE Features
x_title_features = pickle.load(open("video_titles_features_200k.p", "rb"))
x_title_all = x_title_features.todense()
x_title_all = x_title_all[:num_examples]

# Get STYLE Features (Style features characterize the use of exaggeration, forward reference,
# 5w1h, and clickbait words in Video Title)
style_features = data_preparation.find_style_features(x_title_features)
x_style_features = style_features / style_features.max(axis=0)
x_style_features = x_style_features[:num_examples]

# Get VIDEO TAGS Features
x_tags_features = pickle.load(open("video_tags_features_200k.p", "rb"))
x_tags_all = x_tags_features.todense()
x_tags_all = x_tags_all[:num_examples]

# CONCATENATE Video Title, Video Tags, and Style features
x_title_all = numpy.hstack([x_title_all, x_tags_all, x_style_features])

# Set THUMBNAIL IMAGE Shape
thumbnail_rows = 28
thumbnail_cols = 28
thumbnail_channels = 3
if K.image_dim_ordering() == 'th':
    thumbnail_image_shape = (thumbnail_channels, thumbnail_rows, thumbnail_cols)
else:
    thumbnail_image_shape = (thumbnail_rows, thumbnail_cols, thumbnail_channels)
img_size_flatten = thumbnail_rows * thumbnail_cols * thumbnail_channels

# Get THUMBNAILS features
x_images_all = pickle.load(open("video_thumbnail_features_200k.p", "rb"))
x_images_all = x_images_all[:num_examples]
x_images_all = x_images_all.astype('float32') / 255.
x_images_all = numpy.reshape(x_images_all, (num_examples, img_size_flatten))

# Get OTHER Features (These include different statistics like view count and
# counters regarding the comments i.e., comments that contain the word bait)
x_other_all = pickle.load(open("video_other_features_200k.p", "rb"))
x_other_all = x_other_all[:num_examples]

# CONCATENATE ALL INPUTS TOGETHER
x_concat = numpy.hstack([x_title_all, x_other_all, x_images_all])

"""
Get VIDEO LABELS (The video labels file is actually a dictionary with video_ids 
as keys and video labels as values). The labels SHOULD NOT BE one-hot encoded.
"""
all_video_labels = pickle.load(open("video_labels_2k.p", "rb"))

"""
Split Video Input Data (Features) and Video Labels to TRAIN and TEST sets while ensuring that
we have the same amount of videos of each class (Clickait, Non-clickbait) in each set 
"""
# Initially get the Indices of Clickbait (True), Non-clickbait (False), and Unknown (unlabaled videos)
groundtruth_column = list() # the list of labels (1.0, 0.0, or 0.5)
for video_id in dataset_video_ids:
    if (video_id in all_video_labels.keys()):
        groundtruth_column.append(float(all_video_labels[video_id]))
    else:
        groundtruth_column.append(0.5)
indices_true = [i for i, x in enumerate(groundtruth_column) if x == 1.0]
indices_false = [i for i, x in enumerate(groundtruth_column) if x == 0.0]
indices_unknown = [i for i, x in enumerate(groundtruth_column) if x == 0.5]

# Shuffle Indices
indices_true = shuffle(indices_true)
indices_false = shuffle(indices_false)
indices_unknown = shuffle(indices_unknown)

# Get the Indices of TRAIN and TEST set while ensuring class examples consistency
i = 0
indices_ordered = list()
for i in range(0, 2000):
    i = 0
    for j in range(0, 1):
        if len(indices_true) > 0:
            indices_ordered.append(indices_true.pop(0))
            i = i + 1
        if len(indices_false) > 0:
            indices_ordered.append(indices_false.pop(0))
            i = i + 1
    for k in range(0, 100 - i):
        if len(indices_unknown) > 0:
            indices_ordered.append(indices_unknown.pop(0))

# Create TRAIN and TEST SETS
x_all_ordered = numpy.take(x_concat, indices_ordered, axis=0)
y_all_ordered = numpy.take(groundtruth_column, indices_ordered, axis=0)
y_all_ordered = numpy.array(y_all_ordered)
y_all_ordered = numpy.reshape(y_all_ordered, (num_examples, 1))

# Append the Labels at the end of the Input (required by our model for later use in loss function)
x_all = numpy.hstack([x_all_ordered, y_all_ordered])

# Create an array with the Ground Truth labels in batches
wanted_labels = [0.0, 1.0]
groundtruth_batch = x_all[numpy.in1d(x_all[:, x_all.shape[1] - 1], wanted_labels)]
groundtruth_batch = groundtruth_batch[:2000]
groundtruth_labels = groundtruth_batch[:, groundtruth_batch.shape[1] - 1]

# Create X_TRAIN and X_TEST
x_train = x_all[20000:num_examples]
x_test = x_all[0:20000]

# Get the VIDEO IDs of the TRAIN SET
video_ids = numpy.take(dataset_video_ids, indices_ordered, axis=0)
# Store VIDEO IDs in a file
pickle.dump(video_ids, open("video_ids_results.p", "wb"))


"""
Set Variational AutoEncoder Model Parameters
"""
# General Parameters
nb_epochs = 1000
batch_size = 1000
latent_dim = 10
dropout_level = 0.5
epsilon_std = 1.0

# CNN Parameters
nb_filters = 64 # Convolutional filters
nb_conv = 3 # Convolution kernel size
cnn_image_output_shape = (batch_size, 64)

# Initial Input dimensions
original_title_dim = x_title_all.shape[1]
original_other_dim = x_other_all.shape[1]
original_merged_input_dim = original_other_dim + original_title_dim

# FUSING NETWORK Intermediate input dimensions
intermediate_title_dim = 300
intermediate_thumbnail_dim = 32
intermediate_other_dim = 7
intermediate_merged_input_dim = intermediate_title_dim + intermediate_other_dim

# Delete the inputs that we no longer use to free some memory
del x_title_all, x_other_all, x_images_all


'''

BUILD VARIATIONAL AUTOENCODER MODEL

'''

"""
Split the Input to the thumbnails input for the CNN and to all other inputs
"""
# Define an input that contains all the Inputs
x_all_placeholder = Input(batch_shape=(batch_size, original_merged_input_dim + img_size_flatten + 1))
x_all_placeholder2 = tf.check_numerics(x_all_placeholder, 'allinput')

# Split the input to each type of input. We are using Lambda layers to avoid the issue with the Keras Tensors
x_all_other_inputs = Lambda(lambda x: x[:, :original_merged_input_dim], name='x_input')(x_all_placeholder)
current_index = original_merged_input_dim

x_thumbnail_input = Lambda(lambda x: x[:, current_index:current_index + img_size_flatten])(x_all_placeholder)

"""
CNN-layered Architecture for processing the Thumbnails Input
"""
# Add all CNN layers
model_thumbnail_input = Reshape(thumbnail_image_shape, name='model_thumbnail_input')(x_thumbnail_input)

model_thumbnail_1 = Convolution2D(thumbnail_channels,
                                  (2, 2),
                                  padding='same',
                                  activation='tanh',
                                  kernel_initializer='glorot_uniform',
                                  name='model_thumbnail_1')(model_thumbnail_input)
model_thumbnail_2 = MaxPooling2D(pool_size=(2, 2))(model_thumbnail_1)

model_thumbnail_3 = Convolution2D(nb_filters,
                                  (2, 2),
                                  padding='same',
                                  activation='tanh',
                                  kernel_initializer='glorot_uniform',
                                  strides=(2, 2),
                                  name='model_thumbnail_3')(model_thumbnail_2)
model_thumbnail_4 = MaxPooling2D(pool_size=(2, 2))(model_thumbnail_3)

model_thumbnail_5 = Convolution2D(nb_filters,
                                  (nb_conv, nb_conv),
                                  padding='same',
                                  activation='tanh',
                                  kernel_initializer='glorot_uniform',
                                  strides=(1, 1),
                                  name='model_thumbnail_5')(model_thumbnail_4)
model_thumbnail_6 = MaxPooling2D(pool_size=(2, 2))(model_thumbnail_5)

model_thumbnail_7 = Convolution2D(nb_filters,
                                  (nb_conv, nb_conv),
                                  padding='same',
                                  activation='tanh',
                                  kernel_initializer='glorot_uniform',
                                  strides=(1, 1),
                                  name='model_thumbnail_8')(model_thumbnail_6)
model_thumbnail_8 = Reshape(cnn_image_output_shape[1:],
                            name='model_thumbnail_8')(model_thumbnail_7)

# Final Layer
model_thumbnail_9 = Dense(intermediate_thumbnail_dim,
                          activation='sigmoid',
                          kernel_initializer='glorot_uniform',
                          name='model_thumbnail_9')(model_thumbnail_8)

"""
Merge everything together to create the input of the FUSING NETWORK
"""
X_merged_inputs = Concatenate(name='X_merged_inputs')([x_all_other_inputs, model_thumbnail_9])

"""
Fusing Network with Dropout
"""
fusing_network = Dense(intermediate_merged_input_dim,
                       activation='tanh',
                       kernel_initializer='glorot_uniform',
                       name='fusing_network')(X_merged_inputs)
dropout_layer = Dropout(dropout_level, name='model_title_2')(fusing_network)

"""
Gating Network
"""
gating_network_1 = Dense(100,
                         activation='tanh',
                         kernel_initializer='glorot_uniform',
                         name='gating_network_1')(dropout_layer)

# Add Gating Network Softmax Layer
gating_network_final = Dense(2,
                             activation='softmax',
                             kernel_initializer='glorot_uniform',
                             name='gating_network_final')(gating_network_1)

"""
Calculate z-mean (μ) and z_log_variance (σ^2) for each one of the 4 Sub-encoders
"""
# Sub-encoder 1
z_mean_1 = Dense(latent_dim, activation='linear', kernel_initializer='glorot_uniform', name='z_mean_1')(dropout_layer)  # mean 1
z_log_var_1 = Dense(latent_dim, activation='linear', kernel_initializer='glorot_uniform', name='z_variance_1')(dropout_layer)  # variance 1

# Sub-encoder 2
z_mean_2 = Dense(latent_dim, activation='linear', kernel_initializer='glorot_uniform', name='z_mean_2')(dropout_layer)  # mean 2
z_log_var_2 = Dense(latent_dim, activation='linear', kernel_initializer='glorot_uniform', name='z_variance_2')(dropout_layer)  # variance 2


def sampling(args):
    """
    Method that implements the re-parameterization trick, it takes in the
    mean μ and standard deviation σ and outputs a sample z from the corresponding Gaussian
    """
    z_mean, z_log_var = args
    epsilons = list()
    for i in range(nb_epochs):
        epsilon = K.random_normal(shape=(batch_size, latent_dim), mean=0., std=epsilon_std)
        epsilons.append(epsilon)
    epsilon2 = tf.stack(epsilons, axis=2)
    epsilon3 = tf.reduce_mean(epsilon2, 2)
    return z_mean + K.exp(z_log_var) * epsilon3

"""
Calculate z for each sub-encoder
"""
z1 = Lambda(sampling, output_shape=(latent_dim,), name='z1')([z_mean_1, z_log_var_1])
z2 = Lambda(sampling, output_shape=(latent_dim,), name='z2')([z_mean_2, z_log_var_2])

"""
Reshape the output of the Gating output in order to get the output for each class and concatenate 
that with the z value of each sub-encoder
"""
def resh_0(inp):
    return tf.reshape(inp[:, 0], [inp.get_shape()[0].value, 1])
def resh_1(inp):
    return tf.reshape(inp[:, 1], [inp.get_shape()[0].value, 1])

gating_reshape_1 = Lambda(resh_0, name='gating_reshape0')(gating_network_final)
gating_reshape_2 = Lambda(resh_1, name='gating_reshape1')(gating_network_final)

# Concatenate each class gating output with the z values of each sub-encoder
def concat_z1(args):
    z_1, gat_resh0 = args
    return z_1 * tf.tile(gat_resh0, [1, latent_dim])
def concat_z2(args):
    z_2, gat_resh1 = args
    return z_2 * tf.tile(gat_resh1, [1, latent_dim])

z1 = Lambda(concat_z1)([z1, gating_reshape_1])
z2 = Lambda(concat_z2)([z2, gating_reshape_2])

"""
Concatenate all z outputs of each sub-encoder together
"""
def concat_zs(args):
    z_1, z_2, gat_resh_0, gat_resh_1 = args
    return z_1 * tf.tile(gat_resh_0, [1, latent_dim]) + z_2 * tf.tile(gat_resh_1, [1, latent_dim])
z_all = Lambda(concat_zs, name='z_all')([z1, z2, gating_reshape_1, gating_reshape_2])

"""
Decoder (reconstructs the initial input). It includes all 4 Subencoders
"""
decoder_h = Dense(intermediate_merged_input_dim,
                  kernel_initializer='glorot_uniform',
                  activation='tanh',
                  name='decoder_h_title')
decoder_mean = Dense(original_merged_input_dim * 2,
                     kernel_initializer='glorot_uniform',
                     activation='linear',
                     name='decoder_mean_title')

# Mean and Variance of Sub-encoder 1
h_decoded_1 = decoder_h(z1)
x_decoded_mean_1 = decoder_mean(h_decoded_1)

# Mean and Variance of Sub-encoder 2
h_decoded_2 = decoder_h(z2)
x_decoded_mean_2 = decoder_mean(h_decoded_2)

"""
Merge them all together to get the decoded output
"""
x_decoded_all = Concatenate(name='x_decoded_all')([x_decoded_mean_1, x_decoded_mean_2])

"""
VAE LOSS FUNCTION
Method that implements the custom Evidence Lower Bound (ELBO) Loss Function for our Variational Autoencoder (VAE) Model
"""
def vae_loss_function(x_all_videosinput_with_labels, x_decoded_all):
    print('Loss function is running...')

    # Get the merged_input for only the videos in the current batch
    x_merged_input = tf.slice(x_all_videosinput_with_labels, [0, 0], [batch_size, original_merged_input_dim])

    """
    Sub-Encoder 1
    """
    # Get the input of this sub-encoder
    x_decoded_input_1 = tf.slice(x_decoded_all, [0, 0], [batch_size, original_merged_input_dim])
    current_index_decoded = original_merged_input_dim
    # Calc Variance
    x_decoded_1_variance = tf.slice(x_decoded_all, [0, current_index_decoded], [batch_size, original_merged_input_dim])
    x_decoded_1_variance = tf.clip_by_value(x_decoded_1_variance, 1e-06, 10)
    # Calc Variance Factor
    variance1_factor = tf.reduce_sum(tf.exp((x_decoded_1_variance))) / batch_size
    # variance1_factor_plus = tf.reduce_sum(x_decoded_1_variance) / batch_size

    """
    Sub-Encoder 2
    """
    # Get the input of this sub-encoder
    current_index_decoded = current_index_decoded + original_merged_input_dim
    x_decoded_input_2 = tf.slice(x_decoded_all, [0, current_index_decoded], [batch_size, original_merged_input_dim])
    current_index_decoded = current_index_decoded + original_merged_input_dim
    # Calc Variance
    x_decoded_2_variance = tf.slice(x_decoded_all, [0, current_index_decoded], [batch_size, original_merged_input_dim])
    x_decoded_2_variance = tf.clip_by_value(x_decoded_2_variance, 1e-06, 10)
    # Calc Variance Factor
    variance2_factor = tf.reduce_sum(tf.exp((x_decoded_2_variance))) / batch_size
    # variance2_factor_plus = tf.reduce_sum(x_decoded_2_variance) / batch_size

    """
    GET LABELS for all the videos in the current batch
    """
    # Get the index of the whole merged input dims with the thumbnail because we want the last digit which is the label
    current_index = original_merged_input_dim + img_size_flatten
    x_groundtruth_labels = tf.slice(x_all_videosinput_with_labels, [0, current_index], [batch_size, 1])

    """
    [1st TERM OF THE EQUATION]
    [KL-DIVERGENCE] Calc the KL-divergence between the distribution q (of the encoder) and the distribution p (of the decoder)
    """
    # hyperparameter γ - used as a regularization constant to ameroliate the overfitting tendency
    gamma_regularization_const = 0.5
    kl_loss_1 = - gamma_regularization_const * K.mean(1 + z_log_var_1 - K.square(z_mean_1) - K.exp(z_log_var_1), axis=-1)
    kl_loss_2 = - gamma_regularization_const * K.mean(1 + z_log_var_2 - K.square(z_mean_2) - K.exp(z_log_var_2), axis=-1)
    # prior = K.mean(- K.log(0.7) * gating_reshape0 - K.log(0.3) * gating_reshape1, axis=-1)

    # Calculate KL-Divergence Error
    KL_DIVERGENCE = kl_loss_1 + kl_loss_2

    """
    [2nd TERM OF THE EQUATION]
    [RECONSTRUCTION ERROR] Calc mean loss of the Decoder (the part of the equation where γ hyper-parameter is used)
    """
    # Calc the loss of the 1st Sub-encoder
    xent_loss_title_1 = objectives.mse(x_merged_input, tf.nn.sigmoid(x_decoded_input_1))
    # Calc the loss of the 2nd Sub-encoder
    xent_loss_title_2 = objectives.mse(x_merged_input, tf.nn.sigmoid(x_decoded_input_2))

    # Calculate Reconstruction Error
    RECONSTRUCTION_ERROR = (gating_reshape_1 * xent_loss_title_1 / variance1_factor) + (gating_reshape_2 * xent_loss_title_2 / variance2_factor)

    """
    [3RD TERM OF THE EQUATION]
    [CLASSIFICATION ERROR] Calc the Classification Error
    """
    # Calculate Ground Truth Penalty that we be used in the calculation of the Classification Error
    # Create a mask for labeled (0.0, 1.0) and unlabeled (0.5) that a label has not been found.
    # TRUE in in positions of the labels array that are the labeled data and FALSE for where the unlabeled are
    goundtruth_boolean_mask = tf.not_equal(x_groundtruth_labels, 0.5)

    # Apply the Ground Truth positions boolean mask on the dataset labels to get the Ground Truth LABELS
    x_groundtruth = tf.boolean_mask(x_groundtruth_labels, goundtruth_boolean_mask)

    # [OLD WAY]
    # # Calculate Ground Truth penalty based on the output of the Gating Network for all the 4 Classes
    # # Gating Network Output Class 2
    # x_gating_2 = tf.boolean_mask(gating_reshape_2, goundtruth_boolean_mask)
    #
    # Calculate Classification Error
    # CLASSIFICATION_ERROR = tf.reduce_sum(tf.square(x_groundtruth - x_gating_2))
    #
    # """
    # Calculate the result of the custom ELBO (Loss Function)
    # """
    # ELBO_LOSS = RECONSTRUCTION_ERROR + KL_DIVERGENCE + CLASSIFICATION_ERROR + variance1_factor_plus + variance2_factor_plus

    # [NEW WAY]
    # Calculate Ground Truth penalty based on the output of the Gating Network for all the 4 Classes
    # Gating Network Output Class 1
    x_gating_1 = tf.boolean_mask(gating_reshape_1, goundtruth_boolean_mask)

    # Gating Network Output Class 2
    x_gating_2 = tf.boolean_mask(gating_reshape_2, goundtruth_boolean_mask)

    # Concatenate all x_gating_N in one tensor
    x_gating_concat = tf.stack(values=[x_gating_1, x_gating_2], axis=1, name='x_gating_concat')

    # One-hot encode all rows of x_groundtruth
    x_groundtruth_int32 = tf.cast(x_groundtruth, tf.int32)
    x_groundtruth_one_hot = tf.one_hot(indices=x_groundtruth_int32,
                                       depth=4,
                                       on_value=1.,
                                       off_value=0.,
                                       axis=-1,
                                       name='x_groundtruth_one_hot')

    # Calculate Classification Error using Categorical Cross-entropy Formula (TensorFlow Softmax Cross-entropy)
    CLASSIFICATION_ERROR = tf.losses.softmax_cross_entropy(onehot_labels=x_groundtruth_one_hot,
                                                           logits=tf.log(x_gating_concat),
                                                           weights=1.0,
                                                           label_smoothing=0,
                                                           scope=None)
    """
    Calculate the result of the custom ELBO (Loss Function)
    """
    ELBO_LOSS = RECONSTRUCTION_ERROR + KL_DIVERGENCE + CLASSIFICATION_ERROR

    return ELBO_LOSS

"""
Choose an Optimizer for the VAE Model
"""
optimizer = RMSprop(lr=0.001,
                    rho=0.9,
                    epsilon=1e-08,
                    decay=0.0)

"""
Declare the Model's Input and Output
"""
VARIATIONAL_AUTOENCODER_MODEL = Model(x_all_placeholder, x_decoded_all)

# Compile the Model
VARIATIONAL_AUTOENCODER_MODEL.compile(optimizer=optimizer, loss=vae_loss_function)

# Summarize the Model
VARIATIONAL_AUTOENCODER_MODEL.summary()

"""
Get Reference to the different Network outputs. They will be used below when calculating the
Model's Performance Metrics
"""
get_gating_output = K.function([VARIATIONAL_AUTOENCODER_MODEL.layers[0].input, K.learning_phase()], [VARIATIONAL_AUTOENCODER_MODEL.get_layer('gating').output])





"""
TRAIN and TEST the Variational AutoEncoder (VAE) Model
"""
# Get Ground Truth size and train/test proportions
groundtruth_size = groundtruth_batch.shape[0]
groundtruth_train_proportion = int(groundtruth_size * 80 / 100)
groundtruth_test_proportion = groundtruth_size - groundtruth_train_proportion

# Init variables for Train and Test
training_performance_scores, testing_performance_scores = list(), list()

# TRAIN and TEST
for j in range(nb_epochs):
    # Print Epoch counter
    print("--- Epoch = %d/%d" % (j, nb_epochs))

    # TRAIN the model for 1 Epoch
    VARIATIONAL_AUTOENCODER_MODEL.fit(x_train, x_train, batch_size, epochs=1, validation_data=(x_test, x_test))

    """
    TEST the GATING NETWORK on the Labeled Data (Ground Truth).

    We actually make predictions (in batches - according to the batch_size) on the whole Ground Truth dataset using the 
    Gating Network (the output of the Gating Network is the prediction) and then we get the performance metrics for the
    TRAIN and the TEST set by actually splitting the Ground Truth dataset to 80:20 (TRAIN:TEST).
    """
    output, output_prob = list(), list()
    for i in range(int(groundtruth_size / batch_size)):
        batch_output = get_gating_output([groundtruth_batch[i * batch_size:(i * batch_size) + batch_size], 0])[0]
        output_prob.append(batch_output)
        argmax_output = numpy.argmax(batch_output, axis=1)
        output.append(argmax_output)

    groundtruth_all_pred = numpy.hstack(output)
    groundtruth_all_pred_proba = numpy.vstack(output_prob)

    """
    TRAIN PERFORMANCE METRICS
    """
    train_scores = list()

    # Train Accuracy
    train_accuracy = accuracy_score(groundtruth_labels[groundtruth_test_proportion:groundtruth_size],
                                    groundtruth_all_pred[groundtruth_test_proportion:groundtruth_size])
    train_scores.append(train_accuracy)
    # Train Precision
    train_precision = precision_score(groundtruth_labels[groundtruth_test_proportion:groundtruth_size],
                                      groundtruth_all_pred[groundtruth_test_proportion:groundtruth_size],
                                      average='macro')
    train_scores.append(train_precision)
    # Train Recall
    train_recall = recall_score(groundtruth_labels[groundtruth_test_proportion:groundtruth_size],
                                groundtruth_all_pred[groundtruth_test_proportion:groundtruth_size],
                                average='macro')
    train_scores.append(train_recall)
    # Train F1 Score
    train_f1_score = f1_score(groundtruth_labels[groundtruth_test_proportion:groundtruth_size],
                              groundtruth_all_pred[groundtruth_test_proportion:groundtruth_size],
                              average='macro')
    train_scores.append(train_f1_score)

    # Add Performance Metrics to our total metrics array
    training_performance_scores.append(train_scores)

    """
    TEST PERFORMANCE METRICS
    """
    test_scores = list()

    # Test Accuracy
    test_accuracy = accuracy_score(groundtruth_labels[0:groundtruth_test_proportion],
                                   groundtruth_all_pred[0:groundtruth_test_proportion])
    test_scores.append(test_accuracy)
    # Train Precision
    test_precision = precision_score(groundtruth_labels[0:groundtruth_test_proportion],
                                     groundtruth_all_pred[0:groundtruth_test_proportion],
                                     average='macro')
    test_scores.append(test_precision)
    # Train Recall
    test_recall = recall_score(groundtruth_labels[0:groundtruth_test_proportion],
                               groundtruth_all_pred[0:groundtruth_test_proportion],
                               average='macro')
    test_scores.append(test_recall)
    # Train F1 Score
    test_f1_score = f1_score(groundtruth_labels[0:groundtruth_test_proportion],
                             groundtruth_all_pred[0:groundtruth_test_proportion],
                             average='macro')
    test_scores.append(test_f1_score)

    # Add Performance Metrics to our total metrics array
    testing_performance_scores.append(test_scores)

    """
    Print current Epoch's performance metrics
    """
    print("--- Epoch= %d/%d | Acc: %.4f - Precision: %.4f - Recall: %.4f - F1_Score: %.4f," % (j + 1, nb_epochs, train_accuracy, train_precision, train_recall, train_f1_score))
    print("test_Acc: %.4f - test_Precision: %.4f - test_Recall: %.4f - test_F1_Score: %.4f\n" % (test_accuracy, test_precision, test_recall, test_f1_score))

    # Check Test Accuracy Threshold to end Training when reaching that threshold
    threshold = 0.85
    if test_scores[0] >= threshold:
        pickle.dump(groundtruth_labels[0:groundtruth_test_proportion], open(file="groundtruth_labels_test.p", mode="wb"))
        pickle.dump(groundtruth_all_pred_proba[0:groundtruth_test_proportion], open(file="groundtruth_predictions_test.p", mode="wb"))
        break


"""
TEST the GATING NETWORK on all the data (labeled and unlabeled). We actually make predictions on the all the data.
"""
print('\n\n--- TESTing the Model on the WHOLE DATASET...')

# Init variables
gating_result_concat_pred, gating_result_concat_pred_proba = list(), list()

# Init a Progress Bar for the Training Phase
progressBarTestingAll = tqdm(total=int(x_all.shape[0] / batch_size))

# Start Testing
for i in range(int(x_all.shape[0] / batch_size)):
    # Get Predictions probabilities
    batch_output = get_gating_output([x_all[i * batch_size:(i * batch_size) + batch_size], 1])[0]

    # Convert them to class prediction
    gat_output_argmax = numpy.argmax(batch_output, axis=1)

    # Store the predictions/predictions_proba in the array with all the predictions
    gating_result_concat_pred.append(gat_output_argmax)
    gating_result_concat_pred_proba.append(batch_output)

    # Get Prediction counters for each class
    unique, counts = numpy.unique(gat_output_argmax, return_counts=True)
    count_dict = dict(zip(unique, counts))
    try:
        non_clickbait_zeros = count_dict[0]
    except:
        non_clickbait_zeros = 0
    try:
        clickbait_ones = count_dict[1]
    except:
        clickbait_ones = 0
    print('%d. NON-CLICKBAIT= %d, CLICKBAIT= %d' % (i, non_clickbait_zeros, clickbait_ones))

    # Update the Progress Bar
    progressBarTestingAll.update(1)
# Finish Progress Bar
progressBarTestingAll.close()

# Get Total unique predictions for each class

gating_all = numpy.vstack(gating_result_concat_pred)
unique_all, counts_all = numpy.unique(gating_all, return_counts=True)
count_dict_all = dict(zip(unique_all, counts_all))
print('\n\n[Concatenated Results] NON-CLICKBAIT= %d, CLICKBAIT= %d' % (count_dict_all[0], count_dict_all[1]))

"""
Store in a file the predicted labels of all the videos in our Dataset
"""
# Reshape Gating Class Predictions
gating_preds_reshape = numpy.reshape(gating_result_concat_pred, (num_examples,))
pickle.dump(gating_preds_reshape, open("gating_results_predicted_labels.p", "wb"))

# Reshape and store Gating Predicted Probabilities
gating_probs_reshape = numpy.reshape(gating_result_concat_pred_proba, (num_examples, 2))
pickle.dump(gating_probs_reshape, open("gating_results_predicted_probabilities.p", "wb"))

