# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2019 Kostantinos Papadamou, Cyprus University of Technology
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# If you make use of this code in your work, please cite the following paper:
# @inproceedings{zannettou2018good,
#   title={The good, the bad and the bait: Detecting and characterizing clickbait on YouTube},
#   author={Zannettou, Savvas and Chatzis, Sotirios and Papadamou, Kostantinos and Sirivianos, Michael},
#   booktitle={2018 IEEE Security and Privacy Workshops (SPW)},
#   pages={63--69},
#   year={2018},
#   organization={IEEE}
# }


# Import Packages
import os
import numpy
import string
import pickle
from PIL import Image
from pymongo import MongoClient
from sklearn.feature_extraction.text import CountVectorizer
from sklearn import preprocessing


"""
MongoDB Configuration
"""
client = MongoClient('localhost', 27017)
db = client['clickbait_videos']
videos_collection = db.videos

"""
Set Base Directories
"""
THUMBNAILS_BASE_DIR = '/path/to/stored/thumbnails'


def key_exists(element, *keys):
    """
    Check if *keys (nested) exists in `element` (dict).
    :param keys:
    :return: True if key exists, False if not
    """
    if type(element) is not dict:
        raise AttributeError('keys_exists() expects dict as first argument.')
    if len(keys) == 0:
        raise AttributeError('keys_exists() expects at least two arguments, one given.')

    _element = element
    for key in keys:
        try:
            _element = _element[key]
        except KeyError:
            return False
    return True


def get_all_videos_ids_from_collection(collection, number_examples):
    """
    Method that returns a list with all the Video IDs in the given Collection Name
    :param collection_name: the name of the collection to retrive its Video IDs
    :return: a list with all the Video IDs of the given MongoDB Collection
    """
    all_videos = collection.find({})
    all_video_ids = [video['id'] for video in all_videos]
    return all_video_ids[:number_examples]


def is_integer(string):
    """
    Method that checks whether a given string is an Integer
    :param string: the string to check
    :return: True if the string is Integer, False if not
    """
    try:
        int(string)
        return True
    except ValueError:
        return False


def find_style_features(video_titles_raw):
    """
    Method that finds the Style Features of the given Video Titles
    :param video_titles_raw: the list with all the Video Titles to find their style features
    :return: a list with all the Style Features
    """
    all_style_features = list()
    exaggeration_words = ['epic', 'huge', 'very', 'extreme', 'enormous', 'wild']
    forward_reference = ['he', 'she', 'you', 'this', 'these', 'here', 'do']
    contains5w1h = ['which', 'who', 'when', 'why', 'what', 'how']
    baity = ['click', 'exclusive', 'believe', 'happens next', 'you know']
    for title in video_titles_raw:
        title_style_features = list()
        # Count Exclamation Marks
        exclamation_marks = title.count('!')
        # Count Question Marls
        question_marks = title.count('?')
        # Count number of words in Title
        number_words = len(title.split())

        # Count # of Upper words in Title
        count_upper = 0
        words = title.split()
        for word in words:
            if word.isupper():
                count_upper += 1

        # Count Exaggeration, forward reference, etc. words in Title
        exaggeration_count = 0
        forward_reference_count = 0
        count_5w1h = 0
        count_numbers = 0
        baity_count = 0
        for word in words:
            if word.lower() in exaggeration_words:
                exaggeration_count += 1
            if word.lower() in forward_reference:
                forward_reference_count += 1
            if word.lower() in contains5w1h:
                count_5w1h += 1
            if is_integer(word):
                count_numbers += 1
            if word.lower() in baity:
                baity_count += 1

        # Count Punctuation Marks in Title
        punc_counter = 0
        for s in string.punctuation:
            punc_counter += title.count(s)

        # Create a list with all the Style Features of the current Video Title
        title_style_features.append(exclamation_marks)
        title_style_features.append(question_marks)
        title_style_features.append(number_words)
        title_style_features.append(count_upper)
        title_style_features.append(punc_counter)
        title_style_features.append(exaggeration_count)
        title_style_features.append(forward_reference_count)
        title_style_features.append(count_5w1h)
        title_style_features.append(count_numbers)
        title_style_features.append(baity_count)

        # Add this to the list with all the Style Features
        all_style_features.append(title_style_features)

    return numpy.asarray(all_style_features, dtype='float64')


def get_videos_other_features(all_videos):
    """
    Method that retrieves from mongodb the various other data that will be inputted
    to our ML model. For example we add different statistics like view count and
    counters regarding the comments i.e., comments that contain the word bait
    :param all_videos: the list of videos to get their Other Features
    :return: a list of Videos' Other Features
    """
    results = []
    for v in all_videos:
        video = videos_collection.find_one({'id': v})
        try:
            view_count = video['statistics']['viewCount']
        except KeyError:
            view_count = 0
            pass
        try:
            like_count = video['statistics']['likeCount']
        except KeyError:
            like_count = 0
            pass
        try:
            comment_count = video['statistics']['commentCount']
        except KeyError:
            comment_count = 0
            pass
        try:
            favorite_count = video['statistics']['favoriteCount']
        except KeyError:
            favorite_count = 0
            pass
        try:
            dislike_count = video['statistics']['dislikeCount']
        except KeyError:
            dislike_count = 0
            pass
        try:
            misleading_comments_count = len(video['misleadingComments'])
        except KeyError:
            misleading_comments_count = 0
            pass
        try:
            bait_comments_count = len(video['baitComments'])
        except KeyError:
            bait_comments_count = 0
            pass
        try:
            thumbnail_comments_count = len(video['thumbnailComments'])
        except KeyError:
            thumbnail_comments_count = 0
            pass
        try:
            clickbait_comments_count = len(video['clickbaitComments'])
        except KeyError:
            clickbait_comments_count = 0
            pass
        try:
            deceptive_comments_count = len(video['deceptiveComments'])
        except KeyError:
            deceptive_comments_count = 0
            pass
        try:
            deceiving_comments_count = len(video['deceivingComments'])
        except KeyError:
            deceiving_comments_count = 0
            pass
        try:
            clicked_comments_count = len(video['clickedComments'])
        except KeyError:
            clicked_comments_count = 0
            pass
        try:
            flagged_comments_count = len(video['flaggedComments'])
        except KeyError:
            flagged_comments_count = 0
            pass
        try:
            title_comments_count = len(video['titleComments'])
        except KeyError:
            title_comments_count = 0
            pass
        video_result = []
        video_result.append(view_count)
        video_result.append(like_count)
        video_result.append(comment_count)
        video_result.append(favorite_count)
        video_result.append(dislike_count)
        video_result.append(misleading_comments_count)
        video_result.append(bait_comments_count)
        video_result.append(thumbnail_comments_count)
        video_result.append(clickbait_comments_count)
        video_result.append(deceptive_comments_count)
        video_result.append(deceiving_comments_count)
        video_result.append(clicked_comments_count)
        video_result.append(flagged_comments_count)
        video_result.append(title_comments_count)

        results.append(video_result)
    return numpy.asarray(results)


def is_english(languages_dict):
    """
    Method that check the detected Languages dictionary of a Video and filters the non english titles
    :param languages_dict: dictionary with all the detected languages
    :return: True if English is among the detected languages of a Video, False if not
    """
    for language in languages_dict:
        if language['language'] == 'en':
            return True
    return False


def get_video_titles_raw(all_videos):
    """
    Method that returns the video_ids and the titles of each video. Note that it
    filters the language of the title so that we will only acquire the english videos.
    :param number_examples: the maximum number of Video Title to retrieve
    :return: a list with the retrieved Videos TItles
    """
    all_video_titles = list()
    for video_id in all_videos:
        # Get Video's Info
        video_info = videos_collection.find_one({'id': video_id})
        video_title = video_info['snippet']['title']
        if "detectedLanguage" in video_info.keys():
            if is_english(video_info['detectedLanguage']):
                all_video_titles.append(video_title)
    return all_video_titles


def get_video_tags_raw(all_videos):
    """
    Method that retrieves a list with all the Video Tags of the given Videos
    :param all_videos: a list with all the videos to retrieve their video tags
    :return:
    """
    all_video_tags = list()
    for video_id in all_videos:
        video_info = videos_collection.find_one({'id': video_id})
        if key_exists(video_info, 'snippet', 'tags'):
            all_video_tags.append(video_info['snippet']['tags'])
    return all_video_tags


def get_video_thumbnails_features(all_videos, thumbnail_image_shape):
    """
    Method that get a list of urls and retrieves the images, then converts them to
    a numpy array. It returns a numpy array that contains the RGB image data for
    all the images that were requested via the list of urls.
    Note that all Thumbnails are stored in this way:
    <thumbnails_directory>/<video_id>/<video_id>.jpg

    :param video_ids: the videos to retrieve their Thumbnail Image Data
    :return: a numpy array that contains the RGB image data for
    all the images that were requested via the list of urls
    """
    image_list = list()
    i = 0
    for video_id in all_videos:
        if os.path.isfile(THUMBNAILS_BASE_DIR + '/' + video_id + '/' + video_id + '.jpg'):
            with Image.open(THUMBNAILS_BASE_DIR + '/' + video_id + '/' + video_id + '.jpg') as image:
                image2 = image.resize(thumbnail_image_shape, Image.ANTIALIAS)
                image_matrix = numpy.asarray(image2, dtype='float32')
                image_list.append(image_matrix)
                i += 1
    return numpy.asarray(image_list)


def get_video_ids_labels_mapping(all_videos):
    """
    Method that retrieves from MongoDB all the labels for the given Video IDs
    :param video_ids: the Video IDs to retrieve their labels
    :return: a dict with all the Video IDs <-> Labels mappings
    """
    all_video_labels = dict()
    for video_id in all_videos:
        # Get Video Info
        video = videos_collection.find_one({'id': video_id}, {'id': 1, 'clickbaitClassification': 1})
        # Add VideoID <-> Label Mapping to the result
        all_video_labels[video_id] = video['clickbaitClassification']['clickbait']
    return all_video_labels


if __name__ == "__main__":
    """
    Perform Data Preparation by dumping all required data for Training the Variational AutoEncoder in files 
    so that we can later use them to Train and Test our Model 
    """
    # Set the number of Videos to consider
    number_examples = 200000 # Depending on the size of your dataset

    # Get Video IDs and Video Titles raw
    dataset_video_ids = get_all_videos_ids_from_collection(collection=videos_collection, number_examples=number_examples)
    video_titles_raw = get_video_titles_raw(all_videos=dataset_video_ids)

    # Store VIDEO IDs in file
    pickle.dump(dataset_video_ids, open(file="all_video_ids_200k.p", mode="wb"))

    # Generate VIDEO TITLE FEATURES and store them in file
    video_title_vectorizer = CountVectorizer(strip_accents='ascii', max_features=1000, binary=True)
    video_titles_features = video_title_vectorizer.fit_transform(video_titles_raw)
    pickle.dump(video_titles_features, open(file="video_titles_features_200k.p", mode="wb"))
    del video_titles_raw, video_titles_features

    # Generate VIDEO OTHER FEATURES and store them in a file
    video_other_features = get_videos_other_features(all_videos=dataset_video_ids)
    video_other_features = preprocessing.normalize(video_other_features, axis=0).astype("float32")
    pickle.dump(video_other_features, open(file="video_other_features_200k.p", mode="wb"))
    del video_other_features

    # Generate VIDEO TAGS Features
    video_tags_raw = get_video_tags_raw(all_videos=dataset_video_ids)
    video_tags_vectorizer = CountVectorizer(strip_accents='ascii', max_features=100, binary=True)
    video_tags_features = video_tags_vectorizer.fit_transform(video_tags_raw)
    pickle.dump(video_titles_features, open(file="video_tags_features_200k.p", mode="wb"))
    del video_tags_raw, video_tags_features

    # Generate VIDEO THUMBNAILS Features and store them in a file
    thumbnail_rows = 28
    thumbnail_cols = 28
    thumbnail_channels = 3
    thumbnail_image_shape = (thumbnail_rows, thumbnail_cols, thumbnail_channels)

    video_thumbnails_features = get_video_thumbnails_features(all_videos=dataset_video_ids, thumbnail_image_shape=thumbnail_image_shape)
    pickle.dump(video_thumbnails_features, open(file="video_thumbnail_features_200k.p", mode="wb"))
    del video_thumbnails_features

    # Get VIDEO LABELS and store them in a file
    all_video_labels = get_video_ids_labels_mapping(all_videos=dataset_video_ids)
    pickle.dump(all_video_labels, open(file="video_labels_2k.p", mode="wb"))
    del all_video_labels

    del dataset_video_ids
